from __future__ import annotations
from typing import Iterable


class Vec3i:
    '''Vector of three integers. Can represent position of a block.'''

    def __init__(self, x: int, y: int, z: int) -> None:
        self._x = x
        self._y = y
        self._z = z

    @classmethod
    def from_iterable(cls, iterable: Iterable[int]) -> Vec3i:
        '''Create from triple-element list, tuple or other iterable'''
        it = iter(iterable)
        x = next(it)
        y = next(it)
        z = next(it)
        return Vec3i(x, y, z)

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def z(self):
        return self._z



from __future__ import annotations
import contextlib
import io
from typing import Any, Optional, TextIO, cast
import asyncio
import itertools
import sys

import colorama

from .task import TaskManager


def cli_init() -> None:
    colorama.init()


def endswith_line_break(s: str) -> bool:
    return s.endswith(('\r', '\n'))


class WaitIOWrapper:
    def __init__(self, inner: TextIO, message: str = '') -> None:
        self._inner = inner
        self._buffer = io.StringIO()
        self._last_message = message
        self.message = message

    def clear_message(self) -> None:
        # add more spaces to clear ^C
        print('\r', ' '*len(self._last_message), end='\r', file=self._inner)

    def print_message(self) -> None:
        if len(self._last_message) - len(self._message):
            self.clear_message()
        print('\r' + self._message, end='', file=self._inner)
        self._last_message = self._message

    @property
    def message(self) -> str:
        return self._message

    @message.setter
    def message(self, message: str) -> None:
        self._message = message
        self.print_message()

    def __call__(self, message: str) -> None:
        self.message = message

    def write(self, s: str) -> int:
        if s == '':
            return 0
        lines = s.splitlines(True)
        if len(lines) > 1 or endswith_line_break(lines[0]):
            lines[0] = self._buffer.getvalue() + lines[0]
            self._buffer = io.StringIO()
            self.clear_message()
            if endswith_line_break(lines[-1]):
                self._inner.write(''.join(map(lambda s: '  ' + s, lines)))
            else:
                self._inner.write(''.join(map(lambda s: '  ' + s, lines[:-1])))
                self._buffer.write(lines[-1])
            self.print_message()
        else:
            self._buffer.write(s)
        return len(s)

    def __getattr__(self, name: str) -> Any:
        return getattr(self._inner, name)


class Wait:
    def __init__(self, message: Optional[str]) -> None:
        self._message = message

    def __call__(self, message: Optional[str]) -> Any:
        self._message = message

    @property
    def message(self) -> Optional[str]:
        return self._message

    @message.setter
    def message(self, message: Optional[str]) -> None:
        self._message = message


SPINNER_ITER = itertools.cycle('⠸⢰⣠⣄⡆⠇⠋⠙')


@contextlib.asynccontextmanager
async def spinner(message: Optional[str] = None, *, task_manager: TaskManager):
    wait_object = Wait(message)
    stdout = sys.stdout
    wrapper = WaitIOWrapper(sys.stdout)
    sys.stdout = cast(TextIO, wrapper)

    async def wait():
        while True:
            if wait_object.message is not None:
                wrapper.message = f'\r{next(SPINNER_ITER)} {wait_object.message}'
            await asyncio.sleep(0.1)

    task = task_manager.create_task(wait())
    try:
        yield wait_object
    finally:
        task.cancel()
        wrapper.clear_message()
        sys.stdout = stdout


def print_colored(color: str, *values: object, sep: Optional[str] = None,
                  end: Optional[str] = None) -> None:
    if len(values) == 0:
        return
    elif len(values) == 1:
        print(color + str(values[0]) + colorama.Fore.RESET, end=end)
    else:
        print(color + str(values[0]), *values[1:-1],
              values[-1] + colorama.Fore.RESET, sep=sep, end=end)


def info(*values: object, sep: Optional[str] = None,
         end: Optional[str] = None) -> None:
    print(*values, sep=sep, end=end)


def warn(*values: object, sep: Optional[str] = None,
         end: Optional[str] = None):
    print_colored(colorama.Fore.YELLOW, *values, sep=sep, end=end)


def error(*values: object, sep: Optional[str] = None,
          end: Optional[str] = None):
    print_colored(colorama.Fore.RED, *values, sep=sep, end=end)



'''redstone-computer-utilities

Simple debug tools for redstone computers.'''

__all__ = ['register_script', 'create_script', 'run',
           'ResponseError', 'ResponseErrors', 'Script', 'Interface', 'Event',
           'Vec3i', 'Interval', 'gametick', 'redstonetick', 'second']

__version__ = '0.2.0'

from .main import register_script, create_script, run, Script, Interface, Event
from .io import ResponseError, ResponseErrors
from .pos import Vec3i
from .interval import Interval, gametick, redstonetick, second



from __future__ import annotations
from collections import deque
from itertools import zip_longest
from traceback import format_exc, format_exception
import asyncio
from typing import (Any, Callable, Coroutine, Iterable, NoReturn, Optional,
                    Type, TypeVar, Union, cast)
from uuid import UUID
import inspect
from inspect import Parameter
try:
    from types import UnionType, NoneType  # type:ignore
except ImportError:  # Python < 3.10
    UnionType = None  # type:ignore
    NoneType = type(None)  # type:ignore

import typing_extensions
from typing_extensions import Protocol
import uvloop

from .interval import Interval
from .timer import Timer
from .task import TaskManager
from .io import JsonRpcIO, MethodNotFoundError, ResponseError, ResponseErrors
from .pos import Vec3i
from .cli import cli_init, info, warn, error, spinner
from .util import (base64_to_int, base64_to_bytes,
                   int_to_base64, bytes_to_base64)


class Event:
    '''Event.'''

    def __init__(self, name: str, param: Any = None) -> None:
        self._name = name
        self._param = param

    def __eq__(self, other: object) -> bool:
        return (isinstance(other, Event)
                and self._name == other._name
                and self._param == other._param)

    def __hash__(self) -> int:
        return hash((self._name, self.serializable_param))

    def __str__(self) -> str:
        if self._param is None:
            return self._name
        return f'{self._name}({self.serializable_param})'

    def __repr__(self) -> str:
        if self._param is None:
            return f'Event({self._name!r})'
        return f'Event({self._name!r}, {self._param!r})'

    @property
    def name(self) -> str:
        return self._name

    @property
    def param(self) -> Any:
        return self._param

    def to_method_name(self, script: Script) -> str:
        if self._param is not None:
            return f'{script.name}_{self._name}_{self.serializable_param}'
        return f'{script.name}_{self._name}'

    @property
    def serializable_param(self) -> Any:
        return self.param


class SimpleEvent(Event):
    def __init__(self, name: str) -> None:
        super().__init__(name, None)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, SimpleEvent) and super().__eq__(other)

    def __hash__(self) -> int:
        return hash((self._name, self.serializable_param))

    def __repr__(self) -> str:
        return f'SimpleEvent({self._name!r})'


class InterfaceEvent(Event):
    def __init__(self, name: str,
                 interface: Optional[Interface] = None):
        super().__init__(name, interface)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, InterfaceEvent) and super().__eq__(other)

    def __hash__(self) -> int:
        return hash((self._name, self.serializable_param))

    def __repr__(self) -> str:
        if self._param is None:
            return f'InterfaceEvent({self._name!r})'
        return f'InterfaceEvent({self._name!r}, {self._param!r})'

    def with_interface(self, interface: Interface
                       ) -> InterfaceEvent:
        return InterfaceEvent(self.name, interface)

    @property
    def serializable_param(self) -> str:
        return self.param.name


class TimedEvent(Event):
    def __init__(self, name: str,
                 interval: Optional[Interval] = None):
        super().__init__(name, interval)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, TimedEvent) and super().__eq__(other)

    def __hash__(self) -> int:
        return hash((self._name, self.serializable_param))

    def __repr__(self) -> str:
        if self._param is None:
            return f'TimedEvent({self._name!r})'
        return f'TimedEvent({self._name!r}, {self._param!r})'

    def with_interval(self, interval: Interval
                      ) -> TimedEvent:
        return TimedEvent(self.name, interval)

    @property
    def serializable_param(self) -> int:
        return self.param.gametick


class Events:
    ON_SCRIPT_REGISTER = SimpleEvent('onScriptRegister')
    ON_SCRIPT_RELOAD = SimpleEvent('onScriptReload')
    ON_SCRIPT_RUN = SimpleEvent('onScriptRun')
    ON_SCRIPT_INVOKE = SimpleEvent('onScriptInvoke')
    ON_GAMETICK_START = SimpleEvent('onGametickStart')
    ON_GAMETICK_END = SimpleEvent('onGametickEnd')
    ON_INTERFACE_NEW = SimpleEvent('onInterfaceNew')
    ON_GAMETICK_START_DELAY = TimedEvent('onGametickStartDelay')
    ON_GAMETICK_END_DELAY = TimedEvent('onGametickEndDelay')
    ON_GAMETICK_START_CLOCK = TimedEvent('onGametickStartClock')
    ON_GAMETICK_END_CLOCK = TimedEvent('onGametickEndClock')
    ON_INTERFACE_UPDATE = InterfaceEvent('onInterfaceUpdate')
    ON_INTERFACE_UPDATE_IMMEDIATE = InterfaceEvent(
        'onInterfaceUpdateImmediate')
    ON_INTERFACE_READ = InterfaceEvent('onInterfaceRead')
    ON_INTERFACE_WRITE = InterfaceEvent('onInterfaceWrite')
    ON_INTERFACE_REMOVE = InterfaceEvent('onInterfaceRemove')


class Interface:
    '''Interface. Only represents a handle that may bound to a script, the
    actual interface is never processed locally.'''

    def __init__(self, name: str,
                 script: Optional[Script] = None) -> None:
        self._name = name
        self._script = script

    @property
    def name(self) -> str:
        return self._name

    @property
    def script(self) -> Optional[Script]:
        return self._script

    def __eq__(self, other: object) -> bool:
        return (isinstance(other, Interface)
                and self._name == other._name)

    def __hash__(self) -> int:
        return hash((self._name,))

    def __str__(self) -> str:
        return f'interface:{self._name}'

    def __repr__(self) -> str:
        if self._script is None:
            return f'Interface({self._name!r})'
        else:
            return f'Interface({self._name!r}, {self._script!r})'

    def with_script(self, script: Script) -> Interface:
        '''Create a new interface handle that bounds to the given script.'''
        return Interface(self._name, script)

    async def read(self) -> int:
        '''Read from the interface.'''
        if self._script is None:
            return 0
        return await self._script.read_interface(self)

    async def read_bytes(self) -> bytes:
        '''Read from the interface.'''
        if self._script is None:
            return b''
        return await self._script.read_interface_bytes(self)

    async def write(self, data: Union[int, bytes]) -> None:
        '''Write to the interface.'''
        if self._script is None:
            return
        await self._script.write_interface(self, data)


def dict_to_type(v: dict[str, str]) -> Type:
    type_ = v['type']
    if type_ == 'interface':
        return Interface
    elif type_ == 'script':
        return Script
    else:  # literal
        return str


def dict_to_arg(v: dict[str, str], script: Script
                ) -> Union[str, Interface, Script]:
    type_, value = v['type'], v['value']
    if type_ == 'interface':
        return Interface(value).with_script(script)
    elif type_ == 'script':
        return Script(value)
    else:  # literal
        return value


def arg_to_dict(v: Union[str, Interface, Script]) -> dict[str, str]:
    if isinstance(v, Interface):
        type_, value = 'interface', v.name
    elif isinstance(v, Script):
        type_, value = 'script', v.name
    else:  # str
        type_, value = 'literal', v
    return {'type': type_, 'value': value}


def is_union(t: Type) -> bool:
    origin = typing_extensions.get_origin(t)
    if origin is Union:
        return True
    elif UnionType is not None and origin is UnionType:
        return True
    else:
        return False


def get_union_args(t: Type) -> tuple[Type, ...]:
    if t == Parameter.empty:
        return ()
    elif is_union(t):
        return typing_extensions.get_args(t)
    else:
        return (t,)


class NoArgCallback(Protocol):
    def __call__(self) -> Coroutine[Any, Any, None]: ...


class InterfaceArgCallback(Protocol):
    def __call__(self, interface: Interface) -> Coroutine[Any, Any, None]: ...


class ScriptArgCallback(Protocol):
    def __call__(self, script: Script) -> Coroutine[Any, Any, None]: ...


class RunCallback(Protocol):
    def __call__(self, uuid: UUID, run_args: list[dict[str, str]]
                 ) -> Coroutine[Any, Any, int]: ...


class InvokeCallback(Protocol):
    def __call__(self, script: Script, args: dict[str, Any]
                 ) -> Coroutine[Any, Any, Any]: ...


AnyCallback = Callable[..., Coroutine]
# there's no way to define a type to constrain the params of main callback :(
MainRetValT = TypeVar('MainRetValT', int, None,
                      Optional[int])
MainCallback = Callable[..., Coroutine[Any, Any, MainRetValT]]
CallbackT = TypeVar('CallbackT', bound=AnyCallback)
StaticCallbackRegisterer = Callable[[CallbackT], CallbackT]
RuntimeCallbackRegisterer = Callable[
    [CallbackT], Coroutine[Any, Any, CallbackT]]
CallbackRegisterer = Callable[[CallbackT], Any]


class Script:
    '''May represents a handle only or a running script with callable apis,
    which can be inspected with the member ``running``. It is recommended to
    use ``create_script`` rather than register manually.'''

    def __init__(self, name: str, description: str = '',
                 permission_level: int = 2) -> None:
        '''Create a new script. It is recommended to use ``create_script``
        rather than register manually.

        :param name: Name of the script, MUST be unique among all registered
        scripts and MUST be a string consists of only letters, numbers
        and underlines.
        :param description: Description of the script, SHOULD be any unicode
        string that can be displayed as plain text in Minecraft.
        :param permission_level: Permission level required to run the script,
        SHOULD be an integer in [2, 4].'''
        self._name: str = name
        self._description: str = description
        self._permission_level: int = permission_level
        self._event_callbacks: dict[
            Event, list[Union[AnyCallback, Timer]]] = {}
        self._main_callbacks: list[MainCallback] = []
        self._event_method_names: dict[str, Event] = {}
        self._io: JsonRpcIO = cast(JsonRpcIO, None)
        self._id_lock: asyncio.Lock = asyncio.Lock()
        self._id_value: int = 0
        self._auth_key: str = ''
        self._detached_tasks: set[asyncio.Task] = set()
        self._detach_task: asyncio.Task = cast(asyncio.Task, None)
        self._detach_event: asyncio.Event = cast(asyncio.Event, None)
        self._task_manager: TaskManager = cast(
            TaskManager, None)
        self._running: bool = False

    def __eq__(self, other: object) -> bool:
        return (isinstance(other, Script)
                and self._name == other._name)

    def __hash__(self) -> int:
        return hash((self._name,))

    def __str__(self) -> str:
        return f'script:{self._name}'

    def __repr__(self) -> str:
        return f'Script({self._name!r}, ...)'

    @property
    def name(self) -> str:
        return self._name

    def _set_internal(self, io: JsonRpcIO, task_manager: TaskManager):
        self._io = io
        self._task_manager = task_manager
        self._detach_event = task_manager.event()
        self._running = True

    @property
    def running(self) -> bool:
        return self._running

    @property
    def _id(self) -> str:
        self._id_value += 1
        return f'c_{self._name}_{self._id_value}'

    async def _call_api(self, method: str, **params: Any) -> Any:
        if params is None:
            params = {}
        if self._auth_key != '':
            params['authKey'] = self._auth_key
        return await self._io.send(method, params, self._id)

    def _is_event_registered(self, event: Event) -> bool:
        return (event in self._event_callbacks
                and len(self._event_callbacks[event]) > 0)

    def _add_event_callback(self, event: Event,
                            callback: Union[AnyCallback, Timer]
                            ) -> Union[AnyCallback, Timer]:
        if event not in self._event_callbacks:
            self._event_callbacks[event] = []
        self._event_callbacks[event].append(callback)
        self._event_method_names[event.to_method_name(self)] = event
        return callback

    async def _register_event(self, event: Event) -> None:
        await self._call_api(
            'registerCallback',
            event={'name': event.name,
                   'param': event.serializable_param},
            callback=event.to_method_name(self))

    async def _register_event_callback(self, event: Event,
                                       callback: Union[AnyCallback, Timer]
                                       ) -> Union[AnyCallback, Timer]:
        if not self._is_event_registered(event):
            await self._register_event(event)
        self._add_event_callback(event, callback)
        return callback

    def _remove_event_callback(self, event: Event) -> None:
        del self._event_callbacks[event]
        del self._event_method_names[event.to_method_name(self)]

    def _remove_callback(self, event: Event,
                         callback: Union[AnyCallback, Timer]) -> None:
        if self._is_event_registered(event):
            self._event_callbacks[event].remove(callback)

    async def _deregister_event(self, event: Event):
        await self._call_api(
            'deregisterCallback',
            event={'name': event.name,
                   'param': event.serializable_param})

    async def _deregister_event_callbacks(self, event: Event) -> None:
        self._remove_event_callback(event)
        await self._deregister_event(event)

    async def _deregister_callback(self, callback: Union[AnyCallback, Timer]
                                   ) -> None:
        for event, callback_list in list(self._event_callbacks.items()):
            if callback in callback_list:
                callback_list.remove(callback)
                if len(callback_list) == 0:
                    await self._deregister_event_callbacks(event)
        if not isinstance(callback, Timer):
            if callback in self._main_callbacks:
                self._main_callbacks.remove(callback)
                if len(self._main_callbacks) == 0:
                    await self._deregister_callback(self._dispatch_main)

    def _callback_registerer(self, event: Event
                             ) -> CallbackRegisterer[AnyCallback]:
        if self._running:
            return cast(RuntimeCallbackRegisterer[AnyCallback],
                        lambda callback: self._register_event_callback(
                            event, callback))
        else:
            return cast(StaticCallbackRegisterer[AnyCallback],
                        lambda callback: self._add_event_callback(
                            event, callback))

    async def _dispatch_request(self, method: str, params: dict[str, Any]
                                ) -> Any:
        if method in self._event_method_names:
            mapped_params = {}
            key_map = {
                'authKey': 'auth_key',
                'runArgs': 'run_args'
            }
            for k, v in params.items():
                if k in key_map:
                    k = key_map[k]
                elif k == 'interface':
                    v = Interface(v, self)
                elif k == 'script':
                    v = Script(v)
                elif k == 'uuid':
                    if v is not None:
                        v = UUID(v)
                mapped_params[k] = v
            event = self._event_method_names[method]

            async def call_suppress(callback, **kwargs) -> Any:
                try:
                    return await callback(**kwargs)
                except ResponseError as exc:
                    if exc.get_id() is None:
                        raise
                    else:
                        # `raise from` will modify the exception
                        raise (ResponseErrors.SCRIPT_INTERNAL_ERROR
                               .with_cause(exc))
                except Exception:  # pylint: disable=broad-except
                    error(f'Error occurred while running event callback '
                          f'{event} of {self}:')
                    error(format_exc())

            result: Any = None
            response_error: Optional[ResponseError] = None
            last_detach_task = self._detach_task
            for callback in self._event_callbacks[event].copy():
                self._detach_event.clear()
                if isinstance(callback, Timer):
                    self._detach_task = callback.task
                    self._remove_callback(event, callback)
                    callback.event.set()
                else:
                    self._detach_task = self._task_manager.create_task(
                        call_suppress(callback, **mapped_params))
                await asyncio.wait(
                    (self._detach_event.wait(), self._detach_task),
                    return_when=asyncio.FIRST_COMPLETED)
                if self._detach_task.done():
                    try:
                        result = self._detach_task.result()
                    except ResponseError as exc:
                        if exc == ResponseErrors.SCRIPT_INTERNAL_ERROR:
                            error(f'Script internal error occurred while '
                                  f'running event callback {event} of {self}:')
                            error(''.join(format_exception(
                                type(exc), exc, exc.__traceback__)))
                        response_error = exc
                else:
                    self._task_manager.add_task(self._detach_task)
            self._detach_task = last_detach_task
            if response_error is not None:
                raise response_error
            return result
        raise MethodNotFoundError()

    async def _register(self) -> None:
        event = Events.ON_SCRIPT_REGISTER
        self._add_event_callback(event, self._on_script_register)
        await self._call_api(
            'registerScript',
            script=self._name,
            description=self._description,
            permissionLevel=self._permission_level,
            callback=event.to_method_name(self))

    async def _on_script_register(self, auth_key: str) -> None:
        self._auth_key = auth_key
        for event in self._event_callbacks:
            if event != Events.ON_SCRIPT_REGISTER:
                await self._register_event(event)

    async def _deregister(self) -> None:
        if self._auth_key != '':
            await self._call_api('deregisterScript')
            self._auth_key = ''

    def _detach(self) -> None:
        self._detach_event.set()

    async def deregister_callback(self, callback: AnyCallback) -> AnyCallback:
        '''Deregister registered event callback.'''
        await self._deregister_callback(callback)
        return callback

    async def list_script(self) -> dict[str, Script]:
        '''List registered scripts.'''
        return {k: Script(k, v['description'], v['permissionLevel'])
                for k, v in (await self._call_api('listScript')).items()}

    async def invoke_script(self, script: Union[str, Script], **kwargs: Any
                            ) -> Any:
        '''Invoke the script.'''
        if isinstance(script, Script):
            script = script.name
        return await self._call_api('invokeScript', script=script, args=kwargs)

    async def list_callback(self, script: Union[str, Script]) -> list[Event]:
        '''(experimental) List registered event callbacks of the script.'''
        if isinstance(script, Script):
            script = script.name
        return [Event(item['name'], item['param']) for item in
                await self._call_api('listCallback', script=script)]

    async def invoke_callback(self, script: Union[str, Script],
                              event: Union[str, Event],
                              args: Optional[dict[str, Any]] = None) -> Any:
        '''(experimental) Invoke the event callback of the script.'''
        if isinstance(script, Script):
            script = script.name
        if isinstance(event, Event):
            event_dict = {'name': event.name,
                          'param': event.serializable_param}
        else:
            event_dict = {'name': event, 'param': None}
        if args is None:
            args = {}
        return await self._call_api('listCallback', script=script,
                                    event=event_dict, args=args)

    async def new_interface(self, name: str, lsb: Union[Iterable[int], Vec3i],
                            increment: Union[Iterable[int], Vec3i], size: int,
                            world: str = 'minecraft:overworld',
                            args: Optional[dict[str, Any]] = None
                            ) -> Interface:
        '''Create an interface.'''
        if not isinstance(lsb, Vec3i):
            lsb = Vec3i.from_iterable(lsb)
        if not isinstance(increment, Vec3i):
            increment = Vec3i.from_iterable(increment)
        if args is None:
            args = {}
        await self._call_api('newInterface', interface=name,
                             world=world, lsb=[lsb.x, lsb.y, lsb.z],
                             increment=[increment.x, increment.y, increment.z],
                             size=size, args=args)
        return Interface(name, self)

    async def remove_interface(self, interface: Union[str, Interface]) -> None:
        '''Remove the interface.'''
        if isinstance(interface, Interface):
            interface = interface.name
        await self._call_api('removeInterface', interface=interface)

    async def list_interface(self) -> dict[str, tuple[Interface, str, Vec3i,
                                                      Vec3i, int]]:
        '''List interfaces. Values of the result is a tuple (interface handle,
        world, lsb, increment, size).'''
        return {k: (Interface(k, self), v['world'],
                    Vec3i.from_iterable(v['lsb']),
                    Vec3i.from_iterable(v['increment']), v['size'])
                for k, v in (await self._call_api('listInterface')).items()}

    async def read_interface(self, interface: Interface) -> int:
        '''Read from the interface.'''
        return base64_to_int(
            await self._call_api('readInterface', interface=interface.name))

    async def read_interface_bytes(self, interface: Interface) -> bytes:
        '''Read from the interface.'''
        return base64_to_bytes(
            await self._call_api('readInterface', interface=interface.name))

    async def write_interface(self, interface: Interface,
                              data: Union[int, bytes]) -> None:
        '''Write to the interface.'''
        if isinstance(data, bytes):
            await self._call_api(
                'writeInterface',
                interface=interface.name,
                data=bytes_to_base64(data))
        elif isinstance(data, int):
            await self._call_api(
                'writeInterface',
                interface=interface.name,
                data=int_to_base64(data))

    async def gametime(self) -> int:
        '''Get current monotonic world time of the overworld in gametick.

        Guaranteed to return the same value as player executes
        `/time query gametime` in the same gametick. Also guaranteed to return
        the same value when called from callbacks of event onGametickStart and
        onGametickEnd in the same gametick.'''
        return await self._call_api('gametime')

    async def list_player(self) -> dict[UUID, tuple[str, int]]:
        '''(experimental) List online players. Values of the result is a tuple
        (name, permission level).'''
        return {UUID(k): (v['name'], v['permissionLevel'])
                for k, v in (await self._call_api('listPlayer')).items()}

    async def info(self, message: str) -> None:
        '''Log the message as information.'''
        await self._call_api('info', message=message)

    async def warn(self, message: str) -> None:
        '''Log the message as warning.'''
        await self._call_api('warn', message=message)

    async def error(self, message: str) -> None:
        '''Log the message as error.'''
        await self._call_api('error', message=message)

    async def send_info(self, uuid: UUID, message: str) -> None:
        '''Send the message as information to the player.'''
        await self._call_api('sendInfo', uuid=str(uuid), message=message)

    async def send_warn(self, uuid: UUID, message: str) -> None:
        '''Send the message as warning to the player.'''
        await self._call_api('sendWarn', uuid=str(uuid), message=message)

    async def send_error(self, uuid: UUID, message: str) -> None:
        '''Send the message as error to the player.'''
        await self._call_api('sendError', uuid=str(uuid), message=message)

    @property
    def on_script_register(self) -> CallbackRegisterer[NoArgCallback]:
        '''Called when the scripts is almost registered.'''
        return self._callback_registerer(Events.ON_SCRIPT_REGISTER)

    @property
    def on_script_reload(self) -> CallbackRegisterer[NoArgCallback]:
        '''Called when `/rcu reload` is executed.'''
        return self._callback_registerer(Events.ON_SCRIPT_RELOAD)

    @property
    def on_script_run(self) -> CallbackRegisterer[RunCallback]:
        '''Called when `/rcu run` is executed.'''
        return self._callback_registerer(Events.ON_SCRIPT_RUN)

    @property
    def on_script_invoke(self) -> CallbackRegisterer[InvokeCallback]:
        '''Called when invoked by another script.'''
        return self._callback_registerer(Events.ON_SCRIPT_INVOKE)

    @property
    def on_gametick_start(self) -> CallbackRegisterer[NoArgCallback]:
        '''Called at the start of every gametick.'''
        return self._callback_registerer(Events.ON_GAMETICK_START)

    @property
    def on_gametick_end(self) -> CallbackRegisterer[NoArgCallback]:
        '''Called at the end of every gametick.'''
        return self._callback_registerer(Events.ON_GAMETICK_END)

    def on_gametick_start_delay(self, interval: Interval
                                ) -> CallbackRegisterer[NoArgCallback]:
        '''Called only once at the start of the gametick after delay
        interval.'''
        return self._callback_registerer(
            Events.ON_GAMETICK_START_DELAY.with_interval(interval))

    def on_gametick_end_delay(self, interval: Interval
                              ) -> CallbackRegisterer[NoArgCallback]:
        '''Called only once at the end of the gametick after delay
        interval.'''
        return self._callback_registerer(
            Events.ON_GAMETICK_END_DELAY.with_interval(interval))

    def on_gametick_start_clock(self, interval: Interval
                                ) -> CallbackRegisterer[NoArgCallback]:
        '''Called at the start of the gametick for every clock cycle.'''
        return self._callback_registerer(
            Events.ON_GAMETICK_START_CLOCK.with_interval(interval))

    def on_gametick_end_clock(self, interval: Interval
                              ) -> CallbackRegisterer[NoArgCallback]:
        '''Called at the end of the gametick for every clock cycle.'''
        return self._callback_registerer(
            Events.ON_GAMETICK_END_CLOCK.with_interval(interval))

    def on_interface_update(self, interface: Interface
                            ) -> CallbackRegisterer[NoArgCallback]:
        '''Called at the end of the gametick when redstone signal received by
        the interface is changed. Guaranteed to be called at most once within
        a gametick at the end of it.'''
        return self._callback_registerer(
            Events.ON_INTERFACE_UPDATE.with_interface(interface))

    def on_interface_update_immediate(self, interface: Interface
                                      ) -> CallbackRegisterer[NoArgCallback]:
        '''Called immediately when redstone signal received by the interface
        is changed. May be triggered multiple times within a single
        gametick.'''
        return self._callback_registerer(
            Events.ON_INTERFACE_UPDATE_IMMEDIATE.with_interface(interface))

    def on_interface_read(self, interface: Interface
                          ) -> CallbackRegisterer[ScriptArgCallback]:
        '''Called before the interface is read by a script.'''
        return self._callback_registerer(
            Events.ON_INTERFACE_READ.with_interface(interface))

    def on_interface_write(self, interface: Interface
                           ) -> CallbackRegisterer[ScriptArgCallback]:
        '''Called before the interface is written by a script.'''
        return self._callback_registerer(
            Events.ON_INTERFACE_WRITE.with_interface(interface))

    def on_interface_new(self, interface: Interface
                         ) -> CallbackRegisterer[InterfaceArgCallback]:
        '''(experimental) Called after an interface is created successfully.'''
        return self._callback_registerer(
            Events.ON_INTERFACE_UPDATE.with_interface(interface))

    def on_interface_remove(self, interface: Interface
                            ) -> CallbackRegisterer[NoArgCallback]:
        '''(experimental) Called before the interface is removed.'''
        return self._callback_registerer(
            Events.ON_INTERFACE_REMOVE.with_interface(interface))

    async def _dispatch_main(self, uuid: Optional[UUID],
                             run_args: list[dict[str, str]]) -> int:
        async def call_suppress(callback: MainCallback, *args: Any) -> int:
            try:
                result = await callback(*args)
            except ResponseError:
                raise
            except Exception:  # pylint: disable=broad-except
                error(f'Error occurred while running main callback of {self}:')
                error(format_exc())
                return 0
            else:
                if isinstance(result, int):
                    return result
                else:
                    return 0

        result: int = 0
        called = False
        response_error: Optional[ResponseError] = None
        last_detach_task = self._detach_task
        for callback in self._main_callbacks.copy():
            signature = inspect.signature(callback)
            args_wanted = list(signature.parameters.values())
            args: list[Union[Optional[UUID], str,
                             Interface, Script]] = []
            should_call = False
            if len(args_wanted) == 0:
                if len(run_args) == 0:
                    should_call = True
                else:
                    continue
            else:
                if args_wanted[0].annotation == Optional[UUID]:
                    args_wanted = args_wanted[1:]
                    args.append(uuid)
                if len(args_wanted) == len(run_args):
                    zipped: Iterable[tuple] = zip(args_wanted, run_args)
                elif args_wanted[-1].kind == Parameter.VAR_POSITIONAL:
                    if len(args_wanted) - 1 > len(run_args):
                        continue
                    zipped = zip_longest(
                        args_wanted, run_args,
                        fillvalue=args_wanted[-1])
                else:
                    continue
                for wanted, provided in zipped:
                    union_args = get_union_args(wanted.annotation)
                    if (len(union_args) == 0
                            or dict_to_type(provided) in union_args):
                        args.append(dict_to_arg(provided, self))
                    else:
                        break
                else:
                    should_call = True
            if not should_call:
                continue
            called = True
            self._detach_event.clear()
            self._detach_task = self._task_manager.create_task(
                call_suppress(callback, *args))
            await asyncio.wait(
                (self._detach_event.wait(), self._detach_task),
                return_when=asyncio.FIRST_COMPLETED)
            if self._detach_task.done():
                try:
                    result += self._detach_task.result()
                except ResponseError as exc:
                    response_error = exc
            else:
                self._task_manager.add_task(self._detach_task)
        if not called:
            raise ResponseErrors.ILLEGAL_ARGUMENT
        self._detach_task = last_detach_task
        if response_error is not None:
            raise response_error
        return result

    @staticmethod
    def _check_main(callback: MainCallback):
        signature = inspect.signature(callback)
        params = list(signature.parameters.values())
        if len(params) > 0:
            param0 = params[0]
            if param0.kind in (Parameter.POSITIONAL_OR_KEYWORD,
                               Parameter.POSITIONAL_ONLY):
                if param0.annotation == UUID:
                    raise TypeError(
                        f'inappropriate parameter {params[0]} for main '
                        f'function, use Optional[UUID] instead')
                elif param0.annotation == Optional[UUID]:
                    params = params[1:]
            for param in params:
                if param.kind in (Parameter.POSITIONAL_OR_KEYWORD,
                                  Parameter.POSITIONAL_ONLY,
                                  Parameter.VAR_POSITIONAL):
                    for anno in get_union_args(param.annotation):
                        if anno not in (str, Interface, Script):
                            raise TypeError(
                                f'Inappropriate parameter {param} for '
                                f'main function')
                else:
                    raise TypeError(
                        f'Inappropriate parameter {param} for main'
                        f'function, use positional argument instead')
        for anno in get_union_args(signature.return_annotation):
            if anno not in (int, None, NoneType, NoReturn):
                raise TypeError(
                    f'Inappropriate return type '
                    f'{signature.return_annotation} for main function')

    @property
    def main(self) -> CallbackRegisterer[MainCallback]:
        '''Called and dispatched when `/rcu run` is executed.'''
        async def runtime_registerer(callback: MainCallback) -> MainCallback:
            Script._check_main(callback)
            if len(self._main_callbacks) == 0:
                await self.on_script_run(self._dispatch_main)
            self._main_callbacks.append(callback)
            return callback

        def static_registerer(callback: MainCallback) -> MainCallback:
            Script._check_main(callback)
            if len(self._main_callbacks) == 0:
                self.on_script_run(self._dispatch_main)
            self._main_callbacks.append(callback)
            return callback

        if self._running:
            return runtime_registerer
        else:
            return static_registerer

    async def wait(self, interval: Interval) -> None:
        '''Suspend the execution until the interval elapses. Guaranteed to
        complete at the start of a gametick.'''
        await self.wait_gametick_start(interval)

    async def wait_gametick_start(self, interval: Interval) -> None:
        '''Suspend the execution until the interval elapses. Guaranteed to
        complete at the start of a gametick.'''
        event = self._task_manager.event()
        timer = Timer(event, self._detach_task)
        await self._register_event_callback(
            Events.ON_GAMETICK_START_DELAY.with_interval(interval), timer)
        self._detach()
        await event.wait()

    async def wait_gametick_end(self, interval: Interval) -> None:
        '''Suspend the execution until the interval elapses. Guaranteed to
        complete at the end of a gametick.'''
        event = self._task_manager.event()
        timer = Timer(event, self._detach_task)
        await self._register_event_callback(
            Events.ON_GAMETICK_END_DELAY.with_interval(interval), timer)
        self._detach()
        await event.wait()


_registered_scripts: list[Script] = []


def register_script(script: Script) -> None:
    '''Register the script. It is recommended to use ``create_script`` rather
    than register manually.'''
    _registered_scripts.append(script)


def create_script(name: str, description: str = '',
                  permission_level: int = 2) -> Script:
    '''Create a new script and register it.

    :param name: Name of the script. MUST be unique among all registered
    scripts and MUST be a string consists of only letters, numbers
    and underlines.
    :param description: Description of the script. SHOULD be any unicode
    string that can be displayed as plain text in Minecraft.
    :param permission_level: Permission level required to run the script.
    SHOULD be an integer >= 2 and <= 4.'''
    script = Script(name, description, permission_level)
    register_script(script)
    return script


def run(host: str = 'localhost', port: int = 37265,
        enable_uvloop=True, enable_builtins=False) -> None:
    '''Try to connect with script server and enter the main loop.'''
    if enable_uvloop:
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
    cli_init()
    info('  Use Ctrl-C to exit')
    if enable_builtins:
        from . import builtins  # pylint: disable=unused-import,import-outside-toplevel
    while True:
        try:
            asyncio.run(run_async(host, port))
        except KeyboardInterrupt:
            break
        except asyncio.IncompleteReadError:
            pass
    info('  Stopped')


class EventTriggeredError(Exception):
    pass


async def run_async(host: str, port: int) -> None:
    async def dispatch_request(method: str, params: dict[str, Any]
                               ) -> Any:
        for script in _registered_scripts:
            try:
                # pylint: disable=protected-access
                return await script._dispatch_request(method, params)
            except MethodNotFoundError:
                continue
        raise MethodNotFoundError()

    task_manager = TaskManager(asyncio.get_running_loop())

    async with spinner(f'Connecting to script server {host}:{port}',
                       task_manager=task_manager):
        while True:
            try:
                io = JsonRpcIO(*await asyncio.open_connection(host, port),
                               dispatch_request, task_manager)
            except OSError:
                await asyncio.sleep(1)
            else:
                break
        info(f'Connected to script server {host}:{port}')

    async def task_added_event_listener() -> NoReturn:
        nonlocal task_manager
        await task_manager.task_added_event.wait()
        raise EventTriggeredError()

    async def watch_dog() -> NoReturn:
        nonlocal task_manager
        while True:
            _done, _pending = await asyncio.wait(
                task_manager.tasks.copy(),
                return_when=asyncio.FIRST_EXCEPTION)
            tasks_new: deque[asyncio.Task] = deque()
            for task in task_manager.tasks:
                try:
                    task.result()
                except asyncio.InvalidStateError:
                    tasks_new.append(task)
                    continue
                except EventTriggeredError:
                    task_manager.task_added_event.clear()
                    tasks_new.append(task_manager.create_task(
                        task_added_event_listener()))
                    continue
                except asyncio.IncompleteReadError:
                    warn('Disconnected')
                    raise
                except ResponseError:
                    pass
            task_manager.tasks.clear()
            task_manager.tasks.extend(tasks_new)

    task_manager.add_coros(task_added_event_listener(), io.run())
    watch_dog_task = task_manager.create_task(watch_dog())

    scripts_len = len(_registered_scripts)
    success_count = 0
    async with spinner(task_manager=task_manager) as wait:
        for i, script in enumerate(_registered_scripts):
            wait(f'Registering script(s) ({i}/{scripts_len})')
            # pylint: disable=protected-access
            script._set_internal(io, task_manager)
            try:
                await script._register()  # pylint: disable=protected-access
            except Exception:  # pylint: disable=broad-except
                error(f'Error occurred while registering {script}')
                error(format_exc())
            else:
                success_count += 1
                info(f'{script} is registered')
        info(f'Registered {success_count}/{scripts_len} script(s)')
    if success_count == 0:
        raise KeyboardInterrupt()

    async with spinner('Running', task_manager=task_manager) as wait:
        await watch_dog_task
        watch_dog_task.result()



from __future__ import annotations
import asyncio
import json
from typing import Any, Callable, Coroutine, NoReturn, Optional

from .task import TaskManager


class ResponseError(Exception):
    def __init__(self, code: int, message: str, id_: Optional[str] = None
                 ) -> None:
        super().__init__()
        self._code = code
        self._message = message
        self._id = id_

    def __eq__(self, other: object) -> bool:
        return (isinstance(other, ResponseError)
                and self._code == other._code)

    def __hash__(self) -> int:
        return hash((self._code,))

    @classmethod
    def from_response(cls, response: dict[str, Any]) -> ResponseError:
        '''Convert json-rpc response dict to exception.'''
        error = response['error']
        return ResponseError(error['code'], error['message'],
                             response['id'])

    def to_response(self, id_: Optional[str] = None) -> dict[str, Any]:
        '''Convert exception to json-rpc response dict.'''
        if id_ is None:
            id_ = self._id
        return {'jsonrpc': '2.0', 'error': {
            'code': self._code, 'message': self._message}, 'id': id_}

    def with_id(self, id_: str) -> ResponseError:
        return ResponseError(self._code, self._message, id_)

    def with_cause(self, cause: Exception) -> ResponseError:
        result = self.copy()
        result.__cause__ = cause
        return result

    def copy(self) -> ResponseError:
        return ResponseError(self._code, self._message, self._id)

    def get_code(self) -> int:
        '''Get code.'''
        return self._code

    def get_message(self) -> str:
        '''Get message.'''
        return self._message

    def get_id(self) -> Optional[str]:
        '''Get id.'''
        return self._id

    def __str__(self) -> str:
        return json.dumps(self.to_response())


class ResponseErrors:
    INVALID_REQUEST = ResponseError(-32600, "Invalid Request")
    METHOD_NOT_FOUND = ResponseError(-32601, "Method not found")
    INVALID_AUTH_KEY = ResponseError(-1, "Invalid authorization key")
    ILLEGAL_NAME = ResponseError(-2, "Illegal name")
    NAME_EXISTS = ResponseError(-3, "Target with the name already exists")
    INVALID_PERMISSION_LEVEL = ResponseError(-4, "Invalid permission level")
    SCRIPT_NOT_FOUND = ResponseError(-5, "Script cannot be found")
    ILLEGAL_ARGUMENT = ResponseError(-6, "Illegal argument")
    SCRIPT_INTERNAL_ERROR = ResponseError(-7, "Script internal error")
    EVENT_NOT_FOUND = ResponseError(-8, "Event cannot be found")
    EVENT_CALLBACK_ALREADY_REGISTERED = ResponseError(
        -9, "Event callback is already registered")
    EVENT_CALLBACK_NOT_REGISTERED = ResponseError(
        -10, "Event callback is not registered")
    ACCESS_DENIED = ResponseError(-11, "Access denied")
    INTERFACE_NOT_FOUND = ResponseError(-12, "Interface cannot be found")
    PLAYER_NOT_FOUND = ResponseError(-13, "Player cannot be found")
    WORLD_NOT_FOUND = ResponseError(-14, "World cannot be found")
    INVALID_SIZE = ResponseError(-15, "Invalid size")
    BLOCK_NOT_TARGET = ResponseError(
        -16, "Non-target block is found in the interface")


class MethodNotFoundError(Exception):
    pass


class JsonRpcIO:
    PARSE_ERROR_RESPONSE = ('{"jsonrpc":"2.0","error":{"code":-32700,'
                            '"message":"Parse error"},"id":null}'
                            ).encode('utf-8')
    INVALID_REQUEST_RESPONSE = ('{"jsonrpc":"2.0","error":{"code":-32600,'
                                '"message":"Invalid Request"},"id": null}'
                                ).encode('utf-8')

    def __init__(self, reader: asyncio.StreamReader,
                 writer: asyncio.StreamWriter,
                 request_handler: Callable[[str, dict[str, Any]],
                                           Coroutine[Any, Any, Any]],
                 task_manager: TaskManager) -> None:
        self._reader = reader
        self._writer = writer
        self._responses: dict[str, dict[str, Any]] = {}
        self._request_handler = request_handler
        self._response_events: dict[str, asyncio.Event] = {}
        self._task_manager = task_manager

    async def _write(self, data: dict[str, Any]) -> None:
        data_bytes = json.dumps(data, ensure_ascii=False, indent=None,
                                separators=(',', ':')).encode('utf-8')
        await self._write_bytes(data_bytes)

    async def _write_bytes(self, data_bytes: bytes) -> None:
        # an OverflowError is raised when length > 65535
        frame = len(data_bytes).to_bytes(2, 'big', signed=False) + data_bytes
        self._writer.write(frame)
        await self._writer.drain()

    async def _dispatch_request(self, request: dict[str, Any]
                                ) -> None:
        try:
            result = await self._request_handler(
                request['method'], request['params'])
        except ResponseError as exc:
            response = exc.to_response(request['id'])
        except MethodNotFoundError:
            response = ResponseErrors.METHOD_NOT_FOUND.with_id(
                request['id']).to_response()
        else:
            response = {'jsonrpc': '2.0', 'result': result,
                        'id': request['id']}
        await self._write(response)

    async def _dispatch_response(self, response: dict[str, Any]
                                 ) -> None:
        id_ = response.get('id')  # default is None
        if id_ is not None and id_ in self._response_events:
            self._responses[response['id']] = response
            self._response_events[id_].set()

    async def _dispatch(self, data: dict[str, Any]) -> None:
        if 'jsonrpc' not in data or data['jsonrpc'] != '2.0':
            await self._write_bytes(JsonRpcIO.INVALID_REQUEST_RESPONSE)
        elif 'method' in data:
            await self._dispatch_request(data)
        elif 'result' in data or 'error' in data:
            await self._dispatch_response(data)
        else:
            await self._write_bytes(JsonRpcIO.INVALID_REQUEST_RESPONSE)

    async def run(self) -> NoReturn:
        '''Enter main loop'''
        while True:
            length_bytes = await self._reader.readexactly(2)
            length = int.from_bytes(length_bytes, 'big', signed=False)
            data_bytes = await self._reader.readexactly(length)
            try:
                data = json.loads(data_bytes.decode('utf-8'))
            except json.JSONDecodeError:
                await self._write_bytes(JsonRpcIO.PARSE_ERROR_RESPONSE)
            else:
                if isinstance(data, dict):
                    self._task_manager.add_coro(self._dispatch(data))
                elif isinstance(data, list):
                    if len(data) == 0:
                        await self._write_bytes(
                            JsonRpcIO.PARSE_ERROR_RESPONSE)
                    for item in data:
                        self._task_manager.add_coro(self._dispatch(item))
                else:
                    await self._write_bytes(JsonRpcIO.PARSE_ERROR_RESPONSE)

    async def send(self, method: str, params: dict[str, Any], id_: str
                   ) -> Any:
        '''Send a request and receive the corresponding response. A
        ResponseError is raised when an error is received. An OverflowError is
        raised when length of json is greater than 65535'''
        self._response_events[id_] = self._task_manager.event()
        await self._write({'jsonrpc': '2.0', 'method': method,
                           'params': params, 'id': id_})
        await self._response_events[id_].wait()
        del self._response_events[id_]
        response = self._responses.pop(id_)
        if 'error' in response:
            raise ResponseError.from_response(response)
        return response['result']



import asyncio
from collections import deque
from typing import Coroutine, Iterable, Deque


class TaskManager:
    def __init__(self, loop: asyncio.AbstractEventLoop) -> None:
        self.loop = loop
        self.tasks: Deque[asyncio.Task] = deque()
        self.task_added_event = self.event()

    def event(self) -> asyncio.Event:
        asyncio.set_event_loop(self.loop)
        return asyncio.Event()

    def add_task(self, task: asyncio.Task) -> asyncio.Task:
        self.tasks.append(task)
        self.task_added_event.set()
        return task

    def add_tasks(self, *tasks: asyncio.Task) -> Iterable[asyncio.Task]:
        self.tasks.extend(tasks)
        self.task_added_event.set()
        return tasks

    def create_task(self, coro: Coroutine) -> asyncio.Task:
        return self.loop.create_task(coro)

    def create_tasks(self, *coros: Coroutine) -> Iterable[asyncio.Task]:
        return [self.loop.create_task(coro) for coro in coros]

    def add_coro(self, coro: Coroutine) -> asyncio.Task:
        return self.add_task(self.create_task(coro))

    def add_coros(self, *coros: Coroutine) -> Iterable[asyncio.Task]:
        return self.add_tasks(*self.create_tasks(*coros))



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['redstone_computer_utilities']

package_data = \
{'': ['*']}

install_requires = \
['colorama>=0.4.5,<0.5.0',
 'typing-extensions>=4.3.0,<5.0.0',
 'uvloop>=0.17.0,<0.18.0']

setup_kwargs = {
    'name': 'redstone-computer-utilities',
    'version': '0.2.1',
    'description': 'Lightweight and Modular Redstone Computer Debugging Tools.',
    'long_description': '<img src="https://cdn.jsdelivr.net/gh/NKID00/redstone-computer-utilities@dev/src/main/resources/assets/rcutil/icon.png" alt="icon" align="right" height="175">\n\n# Redstone Computer Utilities\n\n>  Lightweight and Modular Redstone Computer Debugging Tools.\n\n## Installation\n\nPython 3.7.2 or newer (CPython or PyPy) is required.\n\n```sh\n$ pip install redstone-computer-utilities\n```\n\nor\n\n```toml\n# pyproject.toml\nredstone-computer-utilities = "^0.2.0"\n```\n\n### Check the GitHub repository [NKID00/redstone-computer-utilities](https://github.com/NKID00/redstone-computer-utilities) for details.',
    'author': 'NKID00',
    'author_email': 'nkid00@pm.me',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/NKID00/redstone-computer-utilities',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.2,<4.0.0',
}


setup(**setup_kwargs)



from base64 import b64decode, b64encode
from math import ceil


def base64_to_bytes(v: str) -> bytes:
    return b64decode(v)


def bytes_to_base64(v: bytes) -> str:
    return b64encode(v).decode('ascii')


def base64_to_int(v: str) -> int:
    return int.from_bytes(base64_to_bytes(v), 'little')


def int_to_base64(v: int) -> str:
    if v < 0:
        raise ValueError('integer must be positive, manual convert required')
    return bytes_to_base64(v.to_bytes(
        ceil(v.bit_length() / 8), 'little'))



from fractions import Fraction
from numbers import Rational, Real
from typing import Union


class Interval:
    '''Non-ambiguous interval.'''

    def __init__(self, interval_gametick: int, tps: Union[int, Rational] = 20):
        self._gametick = interval_gametick
        self._tps = tps

    def __eq__(self, other: object) -> bool:
        return (isinstance(other, Interval)
                and self._gametick == other._gametick)

    def __hash__(self) -> int:
        return hash((self._gametick,))

    @property
    def gametick(self) -> int:
        return self._gametick

    @property
    def redstonetick(self) -> Union[int, Rational]:
        # int is not rational to mypy :(
        interval = Fraction(self._gametick, 2)
        if interval.denominator == 1:
            return interval.numerator
        return interval

    @property
    def second(self) -> Union[int, Rational]:
        interval = Fraction(self._gametick, self._tps)
        if interval.denominator == 1:
            return interval.numerator
        return interval

    @property
    def tps(self) -> Union[int, Rational]:
        return self._tps


def gametick(interval: int, tps: Union[int, Rational] = 20) -> Interval:
    '''Non-ambiguous interval.'''
    return Interval(interval, tps)


def redstonetick(interval: Union[int, Real],
                 tps: Union[int, Rational] = 20) -> Interval:
    '''Non-ambiguous interval. Non-integral value will be floored after
    converted to gametick.'''
    return Interval(int(interval * 2), tps)


def second(interval: Union[int, Real],
           tps: Union[int, Rational] = 20) -> Interval:
    '''Non-ambiguous interval. Non-integral value will be floored after
    converted to gametick.'''
    return Interval(int(interval * tps), tps)
